#define Version "v1.60 Aug2002"
/* 
     bestimmen der moeglichen edge-vectors fuer feste vertex-zahl
     edge vectors in upper triangular lattic form (utlf)
                     most compact form (mcf)
     square + SC-lattice + BCC

HISTORY
  v1.2 get_nb() extrem beschleunigt, April 1999 JS 
  v1.3 get_mcf+get_utlf bei option=32 
  v1.4 testsym, in Klassen sortieren, Sym. bestimmen Feb2000
  v1.5 diverse Kleinigkeiten, get_utlf,get_mcf,compare were buggy
  v1.5.1 new test_shell, bit faster
  v1.5.2 Ig = geometric imperfection
  v1.5.3 bibi on bcc (korrekt?)
  v1.5.4 A==B output,u2i,i2u(new indexing), get_utlf using sym A=SB
  v1.5.5 get_utlf now a00<a11<a22
  v1.5.6 Ig works for square ??
  v1.5.7 opt. 32 works with square
  v1.60  calc. shortest edgevectors for hexagonal lattices (not ready)
 
functions: 
   get_mcf  - most compact form
   get_utlf - upper triangular lattice form
   get_vol  - volume (Determinante)
   get_nb   - number of n-th neighbours

 */ 

#include <stdio.h>
#include <stdlib.h>

int vvv=0;	 /* 1: only 8 nn-bcc-neighbours, 2: only bipartit */
int dimension=3; /* in 2d ignore z */
int angle=3;     /* 2D: angle w in 2*M_PI/12 or 30deg; a1=(1,0) a2=(cos(w),sin(w)) */

/* M.Klemm: Symmetrien von Ornamenten und Kristallen, SpringerVerlag 1982 */
/* erzeugende aller Raumgruppen??? 6zahlige Achse fehlt!? */
/* 3D-rotations det(T)=+1 oder als Permutation von sites tabellieren? */ 
int rotate[1+3+2][3][3]=	/* 6*4 symmetrien der 3D-Basis (ex,ey,ez)*/
{
  1, 0, 0,  0, 1, 0,  0, 0, 1, /* C1 = 1 (Schoenfliess & Hermmann-Mauguinsche) */
 -1, 0, 0,  0, 0,-1,  0,-1, 0, /* C2 = 2  um ( 0, 1,-1) */
  0,-1, 0, -1, 0, 0,  0, 0,-1, /* C2 = 2  um ( 1,-1, 0) */
  0, 0,-1,  0,-1, 0, -1, 0, 0, /* C2 = 2  um ( 1, 0,-1) */
  0, 1, 0,  0, 0, 1,  1, 0, 0, /* C3 = 3  um (-1,-1,-1) */
  0, 0, 1,  1, 0, 0,  0, 1, 0, /* C3 = 3  um ( 1, 1, 1)   = R (Klemm9.10) */
};
/* 3D-reflections det(T)=+1 oder als Vectoren \pm 1 */ 
int reflect[4][3][3]={	/* Refl. an .. oder C2-Drehung um Achse */ 
  1, 0, 0,  0, 1, 0,  0, 0, 1,  /* C1 = 1 */
  1, 0, 0,  0,-1, 0,  0, 0,-1,  /* C2 = 2  um ( 1, 0, 0)  = I1 (Klemm9.10) */
 -1, 0, 0,  0, 1, 0,  0, 0,-1,  /* C2 = 2  um ( 0, 1, 0) */
 -1, 0, 0,  0,-1, 0,  0, 0, 1   /* C2 = 2  um ( 0, 0, 1)  = D2 (Klemm9.10) */
};
/*  rot[4..5]*refl[0..3] = C3 um 8 Raumdiagonalen
 *  rot[1..3]*refl[0..3] = C2 um 6 Flaechendiagonalen (doppelt?)
 *  rot[1   ]*refl[2..3] = C4 um 3   
 *  class1=1          l=1 num=1
 *  class2=refl[1..3] l=2 num=3
 *  class3=rot [1..3] l=2 num=6
 *  class4=           l=4 num=6
 *  class5=rot [4..5] l=3 num=8
 */
int cycle[2*24];		/* number of cycles for symmetry P^n=1 */
int class[2*24];		/* class-number */
int b_all[2*24][3][3];		/* alle Punktsymmetrien des kubischen Gitters ? */
 
/* Bravais-lattice */ 
#define SC  0
#define BCC 1
#define FCC 2
int bravais[4][3][3]={
  1, 0, 0,  0, 1, 0,  0, 0, 1,  /* SC */
 -1, 1, 1,  1,-1, 1,  1, 1,-1,  /* bcc vol=4 */
  0, 1, 1,  1, 0, 1,  1, 1, 0,  /* fcc vol=2 fcc=const*bcc^-1 */
  1, 0, 0,  0, 1, 0,  0, 0, 1   /* HCP ??? */
};
/* volume of bravais-unitcells */
int vol_brav[4]={ 1, 4, 2, 1 };	/* const besser -O2 optimierbar */

int product(int a[3][3],int b[3][3],int ab[3][3]){  /* ab=a*b matrixmult */
  int i,j;
  if(ab==a || ab==b){ printf("---error2---\n");exit(1); }
  for(i=0;i<3;i++)for(j=0;j<3;j++){
    ab[i][j]=a[i][0]*b[0][j]+a[i][1]*b[1][j]+a[i][2]*b[2][j];
  }  
  return 0;
}

/* adjunct(utlf) => ltlf  simpler!??? */
/* TransponiertAdjuncte=det(A)*A^(-1); -> Bildung der Inversen Matrix */
int adjunct(int a[3][3],int adjtrans[3][3]){
   if(adjtrans==a){ printf("---error1---\n");exit(1); }
   adjtrans[0][0]=a[1][1]*a[2][2]-a[1][2]*a[2][1];
   adjtrans[0][1]=a[2][1]*a[0][2]-a[2][2]*a[0][1];
   adjtrans[0][2]=a[0][1]*a[1][2]-a[0][2]*a[1][1];
   adjtrans[1][0]=a[1][2]*a[2][0]-a[1][0]*a[2][2];
   adjtrans[1][1]=a[2][2]*a[0][0]-a[2][0]*a[0][2];
   adjtrans[1][2]=a[0][2]*a[1][0]-a[0][0]*a[1][2];
   adjtrans[2][0]=a[1][0]*a[2][1]-a[1][1]*a[2][0];
   adjtrans[2][1]=a[2][0]*a[0][1]-a[2][1]*a[0][0];
   adjtrans[2][2]=a[0][0]*a[1][1]-a[0][1]*a[1][0];
   return 0;
}

int get_vol(int e[3][3]){ int v;		/* volume=Det(edgevec) */
 v=( e[0][0]*(e[1][1]*e[2][2]-e[1][2]*e[2][1])	/* ax(by cz - bz cy) */ 
    +e[0][1]*(e[1][2]*e[2][0]-e[1][0]*e[2][2])	/* ay(bz cx - bx cz) */
    +e[0][2]*(e[1][0]*e[2][1]-e[1][1]*e[2][0]));/* az(bx cy - by cx) */
    return v;}

int q(int r){ return(r*r); } /* x^2 */

/* return square of length (base depends from angle) */
static inline int get_len(int v[3]){
 if (angle==2) return v[0]*v[0]+v[1]*v[1]+v[2]*v[2]+v[0]*v[1]; /*  60deg */
 if (angle==4) return v[0]*v[0]+v[1]*v[1]+v[2]*v[2]-v[0]*v[1]; /* 120deg */
 if (angle==3) return v[0]*v[0]+v[1]*v[1]+v[2]*v[2];           /*  90deg */
  printf("unsupportet angle=%d",angle); 
  exit(1);
}

int sprod(int v[3],int w[3]){ return v[0]*w[0]+v[1]*w[1]+v[2]*w[2];} 
int get_diff(int a[3],int b[3]){ int i,j,ll;
  for(ll=i=0;i<3;i++){j=a[i]-b[i];ll+=j*j;} return ll;}
int get_sum(int a[3],int b[3],int c[3],int fa,int fb,int fc,int r[3]){
  int i,l;for(l=i=0;i<3;i++){r[i]=fa*a[i]+fb*b[i]+fc*c[i];l+=r[i]*r[i];}
  return l;}

/*  -------------------------------------------------------------------------
+++TeX+++

 Beispiel in zwei Dimensionen. $a=(a_x,a_y)$ und $b=(b_x,b_y)$ sind die 
 zwei Vektoren, die die Elementarzelle aufspannen. 
 Sie sind im allgemeinen nicht senkrecht zueinander. 
 Der Vektor $r=(r_x,r_y)$ liegt innerhalb der Elementarzelle, wenn seine
 Komponenten $(r_a,r_b)$ bezueglich der Elementarvektoren zwischen $0$ und $1$
 liegen.
 \begin{eqnarray*}
      r &=& r_a a   + r_b b \\
 %
    a r &=& r_a a^2 + r_b a b \\
    b r &=& r_a a b + r_b b^2 \\
 %
    r_a &=& \frac{(b^2a-(a,b)b)r}{a^2b^2-(a,b)^2} \\
    r_b &=& \frac{(a^2b-(a,b)a)r}{a^2b^2-(a,b)^2} \\
 \end{eqnarray*}
 Grenzfall $(a,b)=0$ liefert bekannte Beziehungen. 
 Im 3 dimensionalen muss entsprechend ein Gleichungssystem 3.\ Ordnung 
 geloest werden.
 \begin{verbatim}
  maple> solve( { ar=ra*aa+rb*ab+rc*ac,\ 
  maple>          br=ra*ab+rb*bb+rc*bc,\
  maple>          cr=ra*ac+rb*bc+rc*cc},{ra,rb,rc});
  {ra= ( + (bb cc - bc bc) ar + (ac bc - cc ab) br + (bc ab - ac bb) cr)/%1,
   rb=-( + (ab cc - bc ac) ar + (ac ac - cc aa) br + (bc aa - ac ab) cr)/%1,
   rc= ( + (bc ab - bb ac) ar + (ab ac - aa bc) br + (aa bb - ab ab) cr)/%1}
  %1 := + ac bc ab + ac bc ab + bb cc aa - bc bc aa - ac ac bb - ab ab cc
  \end{verbatim}
  Auf die Division durch die Determinante kann evl. verzichtet werden, in dem
  die Einheit auf Det(a,b,c) gesetzt wird.
  Hilfreich koennte die Speicherung der Determinante der Einheitszelle und
  der inversen Elementarzelle (normiert mit der Determinante) sein.
---TeX--- ------------------------------------------------------------ */

int inside(int e[3][3],int x[3],int y[3]){
  /* if x not inside of e, y=inside(x),c=location */
  /* (x,y,z) -> vertex */  
  int aa,ab,ac,bb,bc,cc,ar,br,cr,ra,rb,rc,det;
  aa=e[0][0]*e[0][0]+e[0][1]*e[0][1]+e[0][2]*e[0][2];
  ab=e[0][0]*e[1][0]+e[0][1]*e[1][1]+e[0][2]*e[1][2];
  ac=e[0][0]*e[2][0]+e[0][1]*e[2][1]+e[0][2]*e[2][2];
  bb=e[1][0]*e[1][0]+e[1][1]*e[1][1]+e[1][2]*e[1][2];
  bc=e[1][0]*e[2][0]+e[1][1]*e[2][1]+e[1][2]*e[2][2];
  cc=e[2][0]*e[2][0]+e[2][1]*e[2][1]+e[2][2]*e[2][2];
  det= + ac*bc*ab + ac*bc*ab + bb*cc*aa - bc*bc*aa - ac*ac*bb - ab*ab*cc;
  ar=e[0][0]*x[0]   +e[0][1]*x[1]   +e[0][2]*x[2];
  br=e[1][0]*x[0]   +e[1][1]*x[1]   +e[1][2]*x[2];
  cr=e[2][0]*x[0]   +e[2][1]*x[1]   +e[2][2]*x[2];
  ra= ( + (bb*cc - bc*bc)*ar + (ac*bc - cc*ab)*br + (bc*ab - ac*bb)*cr);
  rb=-( + (ab*cc - bc*ac)*ar + (ac*ac - cc*aa)*br + (bc*aa - ac*ab)*cr);
  rc= ( + (bc*ab - bb*ac)*ar + (ab*ac - aa*bc)*br + (aa*bb - ab*ab)*cr);
  /* Matrix-Form: ra=Adj(AA)*ar */
  if(ra<0)ra=-((-ra-1)/det)-1;else ra/=det;
  if(rb<0)rb=-((-rb-1)/det)-1;else rb/=det;
  if(rc<0)rc=-((-rc-1)/det)-1;else rc/=det;
  if(y){
    y[0]=x[0]-ra*e[0][0]-rb*e[1][0]-rc*e[2][0];
    y[1]=x[1]-ra*e[0][1]-rb*e[1][1]-rc*e[2][1];
    y[2]=x[2]-ra*e[0][2]-rb*e[1][2]-rc*e[2][2];
  }
  return ((ra*ra+rb*rb+rc*rc==0)?1:0);
}

int wurzel(int i){ int j;for(j=0;j*j<i;j++);return j; }
/* moegliche geometrische Abstaende im kubischen Gitter, Vielfachheit??? */
/* dist2=r*r, lt: 0=SC 1=BCC 2=FCC */
int test_shell(int dist2,int lt){	/* test dist2=x^2+y^2+z^2 */
  int x,y,z,j,j1,j2,j3,jj[3]={0,0,0}; j=j1=j2=j3=0;
  for(x=0;                x*x<=dist2;x++)
  for(y=0;y<=x &&     y*y+x*x<=dist2;y++)
  for(z=0;z<=y && z*z+y*y+x*x<=dist2;z++)
  if(x*x+y*y+z*z==dist2){		/* SEARCH VEC x>=y>=z */
    /* return 1; */
    if( dimension==3 ) {
      if(x==0 && y==0 && z==0) j= 1; else
      if(x!=0 && y==0 && z==0) j= 6; else
      if(x==y && y!=0 && z==0) j=12; else
      if(x!=y && y!=0 && z==0) j=24; else
      if(x!=y && y==z && z!=0) j=24; else
      if(x==y && y!=z && z!=0) j=24; else
      if(x==y && y==z && z!=0) j= 8; else
      if(x!=y && y!=z && z!=0) j=48;
    }
    if( dimension==2 ) { if (z!=0) continue;
      if(x==0 && y==0) j= 1; else
      if(x!=0 && y==0) j= 4; else
      if(x==y && y!=0) j= 4; else
      if(x!=y && y!=0) j= 8;
    }
    /* SC immer */                        { jj[SC ]+=j; }
    /* BCC alle gerade oder ungerade! */
    if( (1&(x|y|z))==0 || (1&(x&y&z))==1 ){ jj[BCC]+=j; }
    /* FCC alle gerade oder nur 2 ungerade! */
    if( (1&(x|y|z))==0 || (1&(x^y^z))==0 ){ jj[FCC]+=j; }
  }
  return jj[lt];
}

void outm(int m[3][3],int dim){
  int i,j;if(dim>3) dim=3;
  for(i=3-dim;i<3;i++){ 
    for(j=3-dim;j<3;j++)printf(" %2d",m[i][j]);
    /* if(i<2)printf(","); */
  }
}

void m_copy(int m1[3][3],int m2[3][3]){
  int i,j;for(i=0;i<3;i++)for(j=0;j<3;j++)m2[i][j]=m1[i][j];
}

/* vectoren identisch ? */
int m_same(int m1[3][3],int m2[3][3]){
  int i,j,k=0;for(i=0;i<3;i++)for(j=0;j<3;j++)if(m2[i][j]==m1[i][j])k++;
  return ((k==9)?1:0);
}
/* Gitter gleich ? (ohne Symmetrien) */
int m_gleich(int a1[3][3], int a2[3][3]){
    int bb1[3][3],bb2[3][3],i2,v;
    /* compare  A1*A2^-1=int-matrix?  rechts-links Reihenfolge beachten! */
    v=abs(get_vol(a2));  adjunct(a2,bb1); product(a1,bb1,bb2);
    for(i2=0;i2<3*3;i2++)if(abs(bb2[i2/3][i2%3])%(v)!=0) return 0;
    return 1;
}
/* Gitter gleich ? (mit Symmetrien) 1=gleich 0=ungleich */
int compare(int a1[3][3], int a2[3][3]){
    int bb1[3][3],bb2[3][3],bb3[3][3],i1,i2,v;
    /* compare  A1*T_i*A2^-1=int-matrix? */
    /* rechts - links Reihenfolge beachten! */
    v=abs(get_vol(a2)); if( v!=abs(get_vol(a1)) ) return 0;
    adjunct(a2,bb1);
    for(i1=0;i1<6*4;i1++){
       product(rotate[i1/4] ,bb1,bb2);
       product(reflect[i1%4],bb2,bb3);
       product(  a1,         bb3,bb2);
       for(i2=0;i2<3*3;i2++)if(abs(bb2[i2/3][i2%3])%(v)!=0)break;
       if(i2>=3*3) return 1;
    }
    return 0;
}

/* b = sorted mcf = largest..smallest */
int get_sorted(int mcf[3][3],int b[3][3]){
  int i,j,k1;
  for(i=0;i<3;i++) for(j=0;j<3;j++) b[i][j]=mcf[i][j];
  /* tausche zeilen wenn kleiner und multipliziere mit -1 */
  if(dimension>2)
  if( get_len(b[0])<get_len(b[1]) )
  for(j=0;j<3;j++) { k1=b[0][j];b[0][j]=b[1][j];b[1][j]=-k1; }
  if(dimension>2)
  if( get_len(b[0])<get_len(b[2]) )
  for(j=0;j<3;j++) { k1=b[0][j];b[0][j]=b[2][j];b[2][j]=-k1; }
  if( get_len(b[1])<get_len(b[2]) )
  for(j=0;j<3;j++) { k1=b[1][j];b[1][j]=b[2][j];b[2][j]=-k1; }
  return 0;
}
/* xfache Summation von einer Zeile zu anderer Zeile aendert Det=Vol nicht! */
/* get most compact form = shortest possible edgevectors */
/* 4 0 0  2 -4 0  2 2 -6 => L3-L1+L2 => 0 2 6, len(L3-L1)==len(L3) */
int get_mcf(int l[3][3],int mcf[3][3]){
  int i,j,k,fb,fc,flag,r[3],ib,ic,lm;
  for(i=0;i<3;i++) for(j=0;j<3;j++) mcf[i][j]=l[i][j];
  for(;;){
   flag=1;
   for(i=0;i<3;i++){	/* test every vector Li */
     if(i==0)ib=1;else ib=0;	/* idx to other vectors */
     if(i==2)ic=1;else ic=2;
     lm=get_len(mcf[i]);
     for(fb=-1;fb<2;fb++)
     for(fc=-1;fc<2;fc++)	/*  r = 1+Li + fb*Lib + fc*Lic */
     if( lm>get_sum(mcf[i],mcf[ib],mcf[ic],1,fb,fc,r) )
     { for(k=0;k<3;k++) mcf[i][k]=r[k]; flag=0; i=fb=fc=4; }
   }  
   if(flag) break;
  }
#if 0 /* defined(DEBUG) */ 
  if(!m_gleich(l,mcf))         printf("\n ERROR mcf mcf!=L");
  if(get_vol(l)!=get_vol(mcf)) printf("\n ERROR mcf wrong V");
#endif
  return 0;
}
/* xfache Summation von einer Zeile zu anderer Zeile aendert Det=Vol nicht! */
/* get utlf = upper triangular lattice form */
int get_utlf(int l[3][3],int b[3][3]){	/* Gausssches Verfahren, b=utlf */
  int i,j,v;
  v=get_vol(l);
  for(i=0;i<3;i++) for(j=0;j<3;j++) b[i][j]=l[i][j];
  /* tausche zeilen wenn diagonalelement b[i][i]==0 ist */
#if 0
  if(!b[0][0]) for(j=0;j<3;j++) { k1=b[0][j];b[0][j]=b[1][j];b[1][j]=-k1; }
  if(!b[0][0]) for(j=0;j<3;j++) { k1=b[0][j];b[0][j]=b[2][j];b[2][j]=-k1; }
#endif
  b[1][0]=b[2][0]=b[2][1]=0;
  j=1;		/* Brechstangen-Methode */
/*
# sorted by l11.l22.l12.l23.l13 (nach Flynn)
# 1<=l11<=l22<=l33<=v l11*l22*l33=v
# 1<=l11<=v^(1/3) 0<=2*l23<=l33 0<=l13<l33 0<=2*l12<=l22
*/
  for(b[1][1]=1;j==1 && b[1][1]        <=abs(v);b[1][1]++)if(abs(v)%b[1][1]==0)
  for(b[0][0]=1;j==1 && b[0][0]*b[1][1]<=abs(v);b[0][0]++)if(abs(v)%(b[0][0]*b[1][1])==0)
  if(b[0][0] <= b[1][1]) /* Aug00 */
  for(b[2][2]=v/(b[0][0]*b[1][1]),
      b[1][2]=0;j==1 && b[1][2]<=abs(b[2][2])/2;b[1][2]++)
  for(b[0][1]=0;j==1 && b[0][1]<=    b[1][1] /1;b[0][1]++)
  for(b[0][2]=0;j==1 && b[0][2]<=abs(b[2][2])/2;b[0][2]++){
    if(dimension==2 && (b[0][0]!=1 || b[0][1]!=0 || b[0][2]!=0)) continue;
#if 0
    if(m_gleich(l,b)) { j=0; goto l1; } /* without sym */
#else
    if(compare(l,b)) { j=0; goto l1; } /* using sym v1.5.4 */
#endif
  }
  
  if(j) printf("\n ERROR utlf not found");
  
  l1:
  /* testversion */
  if(!j) if(v!=get_vol(b))  printf("\n ERROR utlf wrong V");
#if 0
  if(!j) if(!m_gleich(l,b)) printf("\n ERROR utlf utlf!=L");
#endif
  return 0;
}

/* evl. statt d pointer to matrix nb[0..d], und fuellen LANGSAM!! */
/* number of neighbours at distance r^.5
  faster: use utlf-Quader for i=0..v-1 loop ( get_nb2 Vergleichen!)
 */
int get_nb(int e[3][3],int d,int *nb){
  int x[3],x2[3],mcf[3][3],v,min[3],max[3],i,vv,dd,m,mm,a0,a1,a2;
  /* use faster utlf-variante */ 
  /* if(e[1][0]==0 && e[2][0]==0 && e[2][1]==0)return get_nb2(e,d,nb); */
  v=get_vol(e);vv=0;
  for(i=0;i<=d;i++)nb[i]=0; if(v==0) return 0;
  for(i=0;i<3;i++){	/* generiere einhuellenden Quader min-max */
    min[i]=max[i]=0;
    for(a0=0;a0<2;a0++)
    for(a1=0;a1<2;a1++)
    for(a2=0;a2<2;a2++)
    { dd=a0*e[0][i]+a1*e[1][i]+a2*e[2][i];
      if(dd<min[i]) min[i]=dd;
      if(dd>max[i]) max[i]=dd;
    }
  }
  /* laengste seite suchen, nur -1,0,1 verschieben
     only mcf ????
  */
  get_mcf(e,mcf);
  for(x[0]=min[0];x[0]<=max[0];x[0]++)
  for(x[1]=min[1];x[1]<=max[1];x[1]++)
  for(x[2]=min[2];x[2]<=max[2];x[2]++) /* vector x inside epiped? */
  if(inside(e,x,NULL)) {
    /* suche kuerzesten abstand zu einem der periodischen 0-Punkte */
    /* starte mit 0-Punkt */
    for(i=0;i<3;i++)x2[i]=x[i]; dd=get_len(x2); mm=m=dd;
    for(;;){	/* suche naeherlegenden punkt in umgebung step by step */
       for(a0=-1;a0<=1;a0++)	/* korrekt? */
       for(a1=-1;a1<=1;a1++)
       for(a2=-1;a2<=1;a2++){
         x2[0]+=mcf[0][0]*a0+mcf[1][0]*a1+mcf[2][0]*a2;
         x2[1]+=mcf[0][1]*a0+mcf[1][1]*a1+mcf[2][1]*a2;
         x2[2]+=mcf[0][2]*a0+mcf[1][2]*a1+mcf[2][2]*a2;
         dd=get_len(x2);
         if(dd<m){ m=dd;a0=a1=a2=8;break; } /* gefunden */
         x2[0]-=mcf[0][0]*a0+mcf[1][0]*a1+mcf[2][0]*a2;	/* old vec */
         x2[1]-=mcf[0][1]*a0+mcf[1][1]*a1+mcf[2][1]*a2;
         x2[2]-=mcf[0][2]*a0+mcf[1][2]*a1+mcf[2][2]*a2;
       }
       if( a0<8 ) break;
    }
    if(m<d){ nb[m]++; } vv++;
  }
  if(vv!=v) printf(" warn sum=%d v=%d\n",vv,v);
  return(v);  
}

/* immer noch recht langsam */
int get_nb2(int e[3][3],int d,int *nb){
  int x[3],x2[3],x3[3],x4[3],x5[3],mcf[3][3],v,i,vv,dd,m,a0,a1,a2;
  static int nb2[100];
  /* use faster utlf-variante */ 
  if(e[1][0]!=0 || e[2][0]!=0 || e[2][1]!=0){ printf("error01");exit(2); }
  v=e[0][0]*e[1][1]*e[2][2];vv=0;
  for(i=0;i<=d;i++)nb[i]=0; if(v==0) return 0;
  get_mcf(e,mcf);
  /* alle Plaetze einer Elementarzelle (hier Quader mit Laengen utlf-diag) */
  for(x[0]=0;x[0]<e[0][0];x[0]++)
  for(x[1]=0;x[1]<e[1][1];x[1]++)
  for(x[2]=0;x[2]<e[2][2];x[2]++) /* vector x inside epiped? */
  { 
    /* suche kuerzesten abstand zu einem der periodischen 0-Punkte */
    /* starte mit 0-Punkt */
    x2[0]=x[0];x2[1]=x[1];x2[2]=x[2]; dd=get_len(x2);
    for(m=dd;m!=0;){	/* suche naeherlegenden punkt in umgebung step by step */
       /* 3x3x3=27 Versuche */
       for(a0=-1;a0<=1;a0++){
         x3[0]=x2[0]+mcf[0][0]*a0;
         x3[1]=x2[1]+mcf[0][1]*a0;
         x3[2]=x2[2]+mcf[0][2]*a0;
       for(a1=-1;a1<=1;a1++){
         x4[0]=x3[0]+mcf[1][0]*a1;
         x4[1]=x3[1]+mcf[1][1]*a1;
         x4[2]=x3[2]+mcf[1][2]*a1;
       for(a2=-1;a2<=1;a2++)if(a0|a1|a2){
         x5[0]=x4[0]+mcf[2][0]*a2;
         x5[1]=x4[1]+mcf[2][1]*a2;
         x5[2]=x4[2]+mcf[2][2]*a2;
         dd=x5[0]*x5[0]+x5[1]*x5[1]+x5[2]*x5[2];if(dd>=m)continue;
         m=dd;a0=a1=a2=8;x2[0]=x5[0];x2[1]=x5[1];x2[2]=x5[2]; /* gefunden */
       }}}
       if( a0<8 ) break;
    }
    if(m<d){ nb[m]++; } vv++;	/* shortest distance */
  }
  if(vv!=v) printf(" warn sum=%d v=%d\n",vv,v);
#if 0
  get_nb(e,d,nb2);
  for(i=0;i<=d;i++)if(nb[i]!=nb2[i])printf(" error diff nb\n");
#endif  
  return(v);  
}

int geom_imperfection(int maxdist2,int *nb,int vvv,int lattice,int v){
  int i4,i5,ig,j,i;
  /* the following things are true only for BCC ??? */
  if(vvv)printf(" nb");
  i4=nb[0];
  i5=1;ig=0;	/* i4=num_nb, ig=geometric_imperfection */
  v/=vol_brav[lattice] /* 4 */;
  for(i=0;i<maxdist2-1;i++)
  if( lattice!=BCC || i%4>1 ){ /* only bcc 3,4,7,8,11,12,.. */
    if( (j=test_shell(i+1,lattice)) ){
      i4+=nb[i+1];
      i5+=((j<v-i5)?j:v-i5); if(i5>i4)ig+=i5-i4;
      if(vvv)printf(" %d",nb[i+1]);
      if(i4>=v)break;
    }
  } /* neighbor 8,6,0,12,.. */
  if(vvv)if(i4<v)printf(" ++%d",v-i4);
  return ig;
}

void lattice2d(){
  int e[2][2];           /* utlf ((a1,a2),(0,b2))  , a3=b3=c1=c2=0,c3=1 */
  int a1,a2,b2,i,v,j,bi,n=8;
  e[1][0]=0;
  for(n=8,j=1;n<=16;n++,j=1)
  for(b2=1;b2<=n   ;b2++)
  for(a2=0;a2<=b2/2;a2++)
  for(a1=1;a1<=b2  ;a1++)
  { 
    e[1][1]=b2;
    e[0][1]=a2; e[0][0]=a1;

    v=( e[0][0]*e[1][1]       /*   ax by */ 
       -e[0][1]*e[1][0]);     /* - ay bx */
    if(abs(v)!=n)continue;

    bi=1-(1&(a1+a2));		/* bipartite? */
    if(((a1+a2)&3)==0)bi=2;
    if(!bi)continue;

    if(a2)if(b2%a2==0 && a1>a2) continue;

    for(i=0;i<2;i++)printf(" (%2d %2d)",e[i][0],e[i][1]);
    printf(" N=%2u j=%6u",v,j);j++;
    printf(" bi%u",bi); /* not bipartite */
    printf("\n");
  }
}

void lattice3d(){
  int e[3][3];           /* utlf ((a1,a2,a3),(0,b2,b3),(0,0,c3)) */
  int a1,a2,a3,b2,b3,c3,i,v,bi,j,n=8;
  e[1][0]=e[2][0]=e[2][1]=0;
  for(n=8,j=1;n<=16;n++,j=1)
  for(c3=1;c3<=n   ;c3++)
  for(b3=0;b3<=c3/2;b3++)
  for(b2=1;b2<=c3  ;b2++)
  for(a3=0;a3<=c3/2;a3++)
  for(a2=0;a2<=b2/2;a2++)
  for(a1=1;a1<=b2  ;a1++)
  { 
    e[2][2]=c3;
    e[1][2]=b3; e[1][1]=b2;
    e[0][2]=a3; e[0][1]=a2; e[0][0]=a1;
    v=get_vol(e);
    if(abs(v)!=n)continue;   

    bi=1-(1&((a1+a2+a3)|(b2+b3))); /* bipartite? */
    if( ((a1+a2+a3)&3)==0 && ((b2+b3)&3)==0 && (c3&3)==0 ) bi=2;
    if(!bi)continue;

    for(i=0;i<3;i++)printf(" (%2d %2d %2d)",e[i][0],e[i][1],e[i][2]);
    printf(" N=%2u j=%6u",v,j);j++;
    printf(" bi%u",bi); /* bipartite */
    printf("\n");
  }
}



int testsym(int a1[3][3]){		/* Jan00 */
    int bb1[3][3],bb2[3][3],i1,i2,i3,v;
    int cno[11],vvv=0;
    for(i1=0;i1<11;i1++)cno[i1]=0;	/* classenstaerke */
    /* compare  A1*T_i*A1^-1=int-matrix? */
    /* rechts - links Reihenfolge beachten! */
    v=abs(get_vol(a1));
    if(vvv)printf(" sym=");
    for(i1=0;i1<6*4;i1++){
       adjunct(a1,bb1);
#if 0
       product(rotate[(i1/4)%6] ,bb1,bb2);
       product(reflect[i1%4],bb2,bb1);
#else
       product(b_all[i1],bb1,bb2);m_copy(bb2,bb1);
#endif       
       product(  a1,         bb1,bb2);
       for(i2=0;i2<3*3;i2++)if(abs(bb2[i2/3][i2%3])%(v)!=0)break;
       i3=((i2>=3*3)?class[i1]:0);	/* ermittelte Klasse 0=keine */
       if(vvv)printf("%d",i3);
       if(i3>=0 && i3<11)cno[i3]++;
    }
    printf(" c");for(i1=1;i1<6;i1++) printf("%d",cno[i1]);
    /* c=Teilmenge_von, Dnh c Dnd c Dn, Cnh c Cnv c Cn, Cs c Ci c C, S2n, D c C */
    /* M.Belger+DP.L.Ehrenberg: Theorie u. Anwendung der Symmetriegruppen K5.3 */
    /* class   C2_um_x,y,z  C2_um_xy,... C4_um_x      C3_um_xyz,... */
    /* cycles  2            2            4            3  Zaehligkeit */
    /* Achsen  3            6            6            8  Achsenzahl */
    if(cno[2]==0 && cno[3]==0 && cno[4]==0 && cno[5]==0) printf(" Ci  ");else
    if(cno[2]==1 && cno[3]==0 && cno[4]==0 && cno[5]==0) printf(" C2h ");else
    if(cno[2]==0 && cno[3]==1 && cno[4]==0 && cno[5]==0) printf(" C2h ");else
    if(cno[2]==0 && cno[3]==0 && cno[4]==0 && cno[5]==2) printf(" C3i ");else
    if(cno[2]==1 && cno[3]==0 && cno[4]==2 && cno[5]==0) printf(" C4h ");else
    if(cno[2]==3 && cno[3]==0 && cno[4]==0 && cno[5]==0) printf(" D2h ");else
    if(cno[2]==1 && cno[3]==2 && cno[4]==0 && cno[5]==0) printf(" D2h ");else
    if(cno[2]==0 && cno[3]==3 && cno[4]==0 && cno[5]==2) printf(" D3h ");else
    if(cno[2]==3 && cno[3]==2 && cno[4]==2 && cno[5]==0) printf(" D4h ");else
    if(cno[2]==3 && cno[3]==6 && cno[4]==6 && cno[5]==8) printf(" Oh  ");else
    printf(" ???");
    return 0;
}

/* ----- new general algorithm ----------- */
#define Buflen  400
#define MaxDist2 30	/* =24 max. moegl. Distanz 2er Punkte im period. Gitter */

/* generate all possible utlf-matrizes
   utlf=unit*basis=(n1,n2,n3)*(b1,b2,b3) => utlf*basis^-1=unit 
 */
void latticebcc(){
  int e[3][3];           /* utlf ((a1,a2,a3),(0,b2,b3),(0,0,c3)) */
  int a1,a2,a3,b2,b3,c3,i,i1,i2,i3,i4,vol,bi,j,nb[MaxDist2+1],n,
      buf[Buflen][3][3],bb,bb1[3][3],bb2[3][3],bb3[3][3],k=0,
      lat[3][3],latv,inv[3][3],invv;
  m_copy(bravais[BCC],lat); latv=get_vol(lat); /* =vol_brav[BCC] */
  adjunct( lat       ,inv); invv=get_vol(inv);
  printf("unit vectors: "); outm(lat,3); printf(" vol: %2u\n",latv);
  printf("adj. vectors: "); outm(inv,3); printf(" vol: %2u\n",invv);
  e[1][0]=e[2][0]=e[2][1]=0;
  for(n=9,j=1,bb=0;n<=40;n+=1,j=1,bb=0){ /* Volumen in Elementarzellen */
    k=0; vol=latv*n;
/*
# sorted by l11.l22.l12.l23.l13 (nach Flynn)
# 1<=l11<=l22<=l33<=v l11*l22*l33=v
# 1<=l11<=v^(1/3) 0<=2*l23<=l33 0<=l13<l33 0<=2*l12<=l22
#           b^-1 * l = unit
#  1/vb * adj(b) * l = unit
#         adj(b) * l = Det * unit  =>  adb11*l11=n*vb, adb21*l11=n*vb, etc.
#                                      schon in Schleifen testbar (faster)
*/
#if 1	/* a la Flynn (1,1,1),(0,2,0),(0,0,2) */
  for(a1= 1;a1      <=2*n;a1+=1) if((4*n)%a1==0)
  for(b2=a1;a1*b2   <=2*n;b2+=1) if((4*n)%(a1*b2)==0) if(b2<=(c3=(4*n)/(a1*b2)))
  if((inv[2][0]*c3)%latv==0)
  if((inv[2][1]*c3)%latv==0)
  if((inv[2][2]*c3)%latv==0)
  for(a2= 0; 2*a2<=b2 ;a2+=1) /* v1.5.4 if a1 is odd, a2,a3 must also be odd */
  for(b3= 0; 2*b3<=c3 ;b3+=1) /* if(b3==0 || b3>=b2) */
  if((inv[2][0]*b3+inv[1][0]*b2)%latv==0)
  if((inv[2][1]*b3+inv[1][1]*b2)%latv==0)
  if((inv[2][2]*b3+inv[1][2]*b2)%latv==0)
  for(a3= 0;   a3< c3 ;a3+=1) /* v1.5.4 if a1 is odd, a2,a3 must also be odd */
#else
  for(c3= 1;                   c3      <=2*n;c3+=1) if((4*n)%c3==0) 
  for(b2= 1;         b2<=c3 && c3*b2   <=4*n;b2+=1)
  for(a1=4*n/(c3*b2);a1<=b2 && a1*b2*c3==4*n;a1+=1)
  for(b3= 0; 2*b3<=c3 ;b3+=1) if( b2<=c3-b3 )
  for(a3= 0; 2*a3<=c3 ;a3+=1)
  for(a2= 0; 2*a2<=b2 ;a2+=1) if( a1<=b2-a2 )
#endif  
  {
    e[2][2]=c3;
    e[1][2]=b3; e[1][1]=b2;
    e[0][2]=a3; e[0][1]=a2; e[0][0]=a1;
    
    bi=1-(1&((a1+a2+a3)|(b2+b3)|c3)); /* bipartite? SC  */
    bi=1-(1&((a1|a2|a3)|(b2|b3)|c3)); /* bipartite? BCC */
    if( ((a1+a2+a3)&3)==0 && ((b2+b3)&3)==0 && (c3&3)==0 ) bi=2;
    if(vvv&2)if(!bi)continue; 

#if 0    
    if(b3)if(c3%b3==0 && b2>b3) continue; /* ??? */
    if(a3)if(c3%a3==0 && a1>a3) continue;
    if(a2)if(b2%a2==0 && a2>a3) continue;
#endif

    product(e,inv,bb1);
    for(i2=0;i2<3*3;i2++)if(abs(bb1[i2/3][i2%3])%latv!=0)break;
    if(i2<3*3) continue;
    
    k++;		/* gueltige edge-vec */
 
    get_nb2(e,MaxDist2,nb);		/* SLOW!!! */
    if(vvv&1)if(nb[3]!=8)continue;	/* only 8 nn neighbours */
    
    printf("L");outm(e,3);

    /* compare with buffer, same lattice? A_N*T_i1*A_i^-1=int-matrix? */
    /* rechts - links Reihenfolge beachten! */
    adjunct(e,bb1);		/* invers=adj/volume */
    for(i3=i=0;i<bb && i3==0;i++){ /* oder Sym*e speichern! schneller */
      for(i1=0;i1<6*4;i1++){
#if 0
       product(rotate[ i1/4],bb1,bb2);
       product(reflect[i1%4],bb2,bb3);
#else
       product(b_all[  i1  ],bb1,bb3);
#endif       
       product(       buf[i],bb3,bb2);
       for(i2=0;i2<3*3;i2++)if(abs(bb2[i2/3][i2%3])%vol!=0)break;
       if(i2>=3*3){ i3=i1+1; break; }
    }}
    if(i3){ 
      i3--; printf(" - %3u SYM= %2d\n",i,i3);
      continue;
    } 

    /* store into buffer */
    m_copy(e,buf[bb]);
    /* for(i=0;i<3*3;i++)buf[bb][i/3][i%3]=e[i/3][i%3]; */
    if(bb<Buflen-1)bb++;else printf("Error: BufOV\n");

    printf(" - %3u",j);j++;
    printf(" bi%u",bi); 			/* bipartite */
    testsym(e);
    printf(" i=%3d",k);    
    i4=geom_imperfection(MaxDist2,nb,0,BCC,vol);
    printf(" Ig %2d",i4);
    i4=geom_imperfection(MaxDist2,nb,1,BCC,vol);
    
    printf("\n"); fflush(stdout);
  }
  printf("N= %2d num= %3d tested= %3d\n",n,bb,k);
  }
}

void latticesquare(){
  int e[3][3];           /* utlf ((a1,a2,a3),(0,b2,b3),(0,0,c3)) */
  int a1,a2,a3,b2,b3,c3,i,i1,i2,i3,vol,bi,j,nb[MaxDist2+1],n,
      buf[Buflen][3][3],bb,bb1[3][3],bb2[3][3],k=0;
  e[1][0]=e[2][0]=e[2][1]=0;
  dimension=2;
  for(n=4,j=1,bb=0;n<=64;n+=1,j=1,bb=0){ /* or n as argument */
    printf("N=%d\n",n);k=0;
    for(c3= 1;     c3<=n;c3+=1) if( n%c3==0) 
    for(b2= 1;     b2<=n;b2+=1) if(c3*b2==n)
    for(a1= 1;     a1< 2;a1+=1)
    for(b3= 0; 2*b3<=c3 ;b3+=1) if( b2<=c3-b3 )
    for(a3= 0;     a3<1 ;a3+=1)
    for(a2= 0;     a2<1 ;a2+=1)
    {
      k++;
      e[2][2]=c3;
      e[1][2]=b3; e[1][1]=b2;
      e[0][2]=a3; e[0][1]=a2; e[0][0]=a1;
      
      vol=get_vol(e);		/* on utlf simply a1*b2*c3 */
      if(vol!=n)continue;
      
      bi=1-(1&((b2+b3)|c3)); /* bipartite? */
      if(vvv&2)if(!bi)continue; 
      if( (((a1|a2|a3)&1)==0 || dimension==2)
       && ((b2|b3)&1)==0 && (c3&1)==0  ) bi=2;
      
  //    if(b3)if(c3%b3==0 && b2>b3) continue;

      get_nb(e,MaxDist2,nb);		/* SLOW!!! */
      if(vvv&1)if(nb[1]!=4)continue;		/* only 4 nn neighbours */
      
      outm(e,2);

      /* compare with buffer, same lattice? A_N*T_i1*A_i^-1=int-matrix? */
      /* rechts - links Reihenfolge beachten! */
      for(i3=i=0;i<bb && i3==0;i++)for(i1=0;i1<6*4;i1++){
         adjunct(buf[i],bb1);
         product(rotate[i1/4] ,bb1,bb2);
         product(reflect[i1%4],bb2,bb1);
         product(  e,          bb1,bb2);
         for(i2=0;i2<3*3;i2++)if(abs(bb2[i2/3][i2%3])%vol!=0)break;
         if(i2>=3*3){ i3=i1+1; break; }
      }
      if(i3){ 
        i3--; printf(" - %2u SYM=%d\n",i,i3);
        continue;
      } 

      /* store into buffer */
      for(i=0;i<3*3;i++)buf[bb][i/3][i%3]=e[i/3][i%3];
      if(bb<Buflen-1)bb++;else printf("Error: BufOV\n");

      printf(" - %2u",j);j++;
      printf(" bi%u",bi); 			/* bipartite */
      testsym(e);
      printf(" i=%3d",k);
      i2=geom_imperfection(MaxDist2,nb,0,SC,vol);
      printf(" Ig %2d",i2);
      i2=geom_imperfection(MaxDist2,nb,1,SC,vol);
      
      printf("\n"); fflush(stdout);
    }
  }
}

void inisym(int vvv){
    int bb1[3][3]={{1,0,0},{0,1,0},{0,0,1}},
        bbi[3][3]={{-1,0,0},{0,-1,0},{0,0,-1}},
        bb2[3][3],bb3[3][3],bb5[3][3],i1,i2,i3,classes=1;
    for(i1=0;i1<2*6*4;i1++){
       product(rotate[(i1/4)%6] ,bb1,bb2);
       product(reflect[i1%4],bb2,bb3);
       if((i1/(6*4))&1)product(bbi,bb3,bb2);else m_copy(bb3,bb2);
       m_copy(bb2,b_all[i1]);
    }
    for(i1=0;i1<2*6*4;i1++){
       class[i1]=classes;
       m_copy(b_all[i1],bb3);
       if(vvv){ printf(" %2u U=",i1); outm(bb3,3); } i3=0;
       if(vvv)for(i2=0;i2<i1;i2++)if(m_same(b_all[i2],bb3)) printf(" = %2d",i2);
       m_copy(bb3,bb2);
       for(i3=1;i3<9;i3++){
         if(m_same(bb2,bb1)) break;
         product(bb3,bb2,bb5);m_copy(bb5,bb2);
       }
       if(vvv)printf("  P=%d",i3); cycle[i1]=i3;	/* Zyklenlaenge */
       if(vvv)printf(" D=%2d",get_vol(bb3));		 /* Determinante */
       if(vvv)printf(" S=%2d",bb3[0][0]+bb3[1][1]+bb3[2][2]);  /* Spur */
/* Aequivalenzklassen bestimmen, A und B sind aequivalent wenn A*C=C*B (alle C)
 * bei Permutationen gehoeren alle Elemente mit gleicher Zyklendarstellung
 * zur selben Klasse (schneller!!!???), (liste der Zyklenlaengen)
 * - mach langsames in einer precompilerphase und binde als static ein?
 */
       for(i3=0;i3<2*24;i3++){
           product(bb3,b_all[i3],bb5);		/* A*C A=i1  C=i3 */
         for(i2=0;i2<i1;i2++){
           product(b_all[i3],b_all[i2],bb2);	/* C*B C=i3  B=i2 */
           if(m_same(bb2,bb5)) { class[i1]=class[i2];i2=i3=2*24; break; }
         }
       }
       if(vvv)printf(" class=%2d\n",class[i1]);
       if(class[i1]==classes) classes++;
    }
}

/*
#   generate index from utlf form,  JS Mai2000 
#   l11 l12 l13      1<=l11<=l22<=l33<=2N, l12<=l22, l13<=l33, l23<=l33
#    0  l22 l23         l11*l22*l33=4N,                        l23<=l33/2
#    0   0  l33 
# base of bcc     (-1,1,1),   (1,-1,1),   (1,1,-1)
#      in utlf  a1=(1,1,1), a2=(0,2,0), a3=(0,0,2)
#         
#      l1 = a (1,1,1) + b (0,2,0) + c (0,0,2)
#      l2 =             d (0,2,0) + e (0,0,2)
#      l3 =                         f (0,0,2)
#
# estimation of upper limit of number of degrees of freedom (integer values)
#  a first estimation would be: 5 variables from 0 to N step 2 => (N/2+1)^5
#  a better estimation is:
# component    values    maxnum_values   expl: N=9 15 32 36  remark
# l33     even 2,4,...,2N      n33=1             1  1  1  1  fixed l33=4N/(l11*l22)
# l13-x   even 0,2,4,..,N      n13=N/2+1         5  8 17 19  if ( l11 even ) x=0 else x=1
# l23     even 0,2,4,..,N      n23=N/2+1         5  8 17 19
# l12-x   even 0,2,4,,(4N)^.5  n12=((4N)^.5)/2+1 4  4  6  7
# l22     even 2,4,..,(4N)^.5  n22=((4N)^.5)/2   3  3  5  6
# l11          1,..,(4N)^(1/3) n11=(4N)^(1/3)    3  3  5  5
#                                          prod=900,,43350,75810
# example: 1 1 7  0 2 4  0 0 18   idx=13
# the indizees should be in same order of range like old indizes
# better algo would use l11*l22*l33=4N max. 10 possibilities for N<64
# x = l11 & 1
# idx=(l13-x)/2 + n13 * (
#      l23   /2 + n23 * (
#     (l12-x)/2 + n12 * (
#     (l22-2)/2 + n22 * (
#     (l11-1)            ))))
# l11= idx/(n13*n23*n12*n22)+1;		# <<< +1 was forggotten v0.1
# x=l11 & 1;
# l22=(idx/(n13*n23*n12) % n22)*2+2;     # % means modulo (rest of division)
# l12=(idx/(n13*n23)     % n12)*2+x;
# l23=(idx/(n13)         % n23)*2;
# l13=(idx               % n13)*2+x;
# l33=4*n/(l11*l22);			# finaly
*/
int u2i(int l[3][3]){
 int n=l[0][0]*l[1][1]*l[2][2]/4,x=l[0][0]&1,
     n13,n23,n12,n22,l13,l23,l12,l22,l11,idx;
 if(dimension==2) return 0;
 if( l[1][0]!=0 || l[2][0]!=0 || l[2][1]!=0 || l[0][0]*l[1][1]*l[2][2]!=4*n
  || l[0][0]>l[1][1] || l[1][1]>l[2][2]
  || (l[0][2] - x) % 2 == 1 || (l[0][1] - x) % 2 == 1
  || (l[1][2])     % 2 == 1 || (l[1][1])     % 2 == 1 ) { fprintf(stderr,"ERR10\n");exit(1); }
 n13=n23=n/2 + 1;
 n12=0;while ( (n12*n12) <= (4*n) ) { n12++; };
 n12=(n12-1)/2 + 1; n22=n12-1;
 l13= ( l[0][2] - x ) / 2;
 l23=   l[1][2]       / 2;
 l12= ( l[0][1] - x ) / 2;
 l22= ( l[1][1] - 2 ) / 2;
 l11=   l[0][0] - 1;
 if ( l13>=n13 || l23>=n23 || l12>=n12 || l22>=n22
   || l13 < 0  || l23 < 0  || l12 < 0  || l22 <  0) {exit(2);}
 idx=l13 + n13 * (l23 + n23 * (l12 + n12 * (l22 + n22 * l11)));
 return idx;
}
int i2u(int n,int idx, int l[3][3]){
 int n13,n23,n12,n22,x,l11,l22,l12,l23,l13,l33;
 n13= n / 2 + 1; n23= n / 2 + 1; 
 n12=0;while (n12*n12 <= 4*n) {n12++;} n12=(n12-1)/2+1;n22=n12-1;
 l11= idx / (n13 * n23 * n12 * n22) + 1; x=l11 & 1;
 l22=(idx / (n13 * n23 * n12) % n22) * 2 + 2;
 l12=(idx / (n13 * n23)       % n12) * 2 + x;
 l23=(idx / (n13)             % n23) * 2    ;
 l13=(idx                     % n13) * 2 + x;
 l33=4 * n / (l11 * l22);
 if ( l11>l22 ) { exit(3); }
 l[0][0]=l11; l[0][1]=l12; l[0][2]=l13;
 l[1][0]=  0; l[1][1]=l22; l[2][2]=l23;
 l[2][0]=  0; l[2][1]=  0; l[2][2]=l33;
 return 0;
}


void help(int r){
 printf(" edgevec " Version "\n"
        " options/arguments:\n"
        "  -m1 : only 8 bcc-nn\n"
        "  -m2 : only bipartit\n"
        "  -m4 : square lattices\n"
        "  -m8 : bcc lattices\n"
        "  -m16 a11 .. a33 b11 .. b33 : compare A,B (3D)\n"
        "  -m16 a11 .. a22 b11 .. b22 : compare A,B (2D)\n"
        "  -m32 a11 .. a33: get_mcf get_utlf get_vol\n"
        "  -m64 : info about 3d pointsymmetries\n"
        "  -a#  : angle (only 2D) between base-vectors in units of 2pi/12 (default: 3)\n"
        "  | grep -v SYM # filter symmetric utlf\n"
        "example: edgvec -a3 -m32  4 0  0 4\n"
       ); 
 exit(r); 
}

int main(int argc,char *argv[]){
    int i,j,d=3,v,an;
    int a[3][3]={{1,0,0},{0,1,0},{0,0,1}},
        b[3][3]={{1,0,0},{0,1,0},{0,0,1}},c[3][3],b2[3][3],
        nb[MaxDist2+1];
    if(argc<2) help(1);
    for (an=1;an<argc;an++) {
      if (argv[an][0]!='-') break; /* first argument (!option) */
      if (argv[an][1]=='m') vvv=  atoi(argv[an]+2);
      if (argv[an][1]=='a') angle=atoi(argv[an]+2);
    }
    setvbuf(stdout,(char *)NULL,_IONBF,0);	/* no buffer */
    if(vvv&16){
      if(argc-an==2*9) dimension=d=3; else
      if(argc-an==2*4) dimension=d=2; else help(2);
      /* printf("# %dd\n",d); */
      for(i=0;i<d*d;i++)a[3-d+i/d][3-d+i%d]=atoi(argv[i+an]);
      for(i=0;i<d*d;i++)b[3-d+i/d][3-d+i%d]=atoi(argv[i+an+d*d]);
      printf(" A=");outm(a,d);
      printf(" B=");outm(b,d);
      if(!compare(a,b)){ printf(" A!=B"); } else 	/*  A*T*A^-1=U  */
      if(m_gleich(a,b)){ printf(" A==B"); } /* v1.5.4 */
      else             { printf(" A=SB"); } /* same under cubic symmtry */ 
    }
    if(vvv&(64+32+4+8))inisym(vvv&64);
    if(vvv&32){
      if(argc-an==9) dimension=d=3; else
      if(argc-an==4) dimension=d=2; else help(3);
      /* printf("# %dd",d); */
      for(i=0;i<d*d;i++)a[3-d+i/d][3-d+i%d]=atoi(argv[i+an]);
      v=j=get_vol(a);   printf(  "# L   ="); outm(a ,d);
      get_mcf(a,b);     printf("\n# mcf ="); outm(b ,d);
      get_sorted(b,b2); printf("\n# sort="); outm(b2,d);
      get_utlf(a,c);    printf("\n# utlf="); outm(c ,d);
      printf("\n# V= %4d L^2=",j);
      for(i=0;i<3;i++)printf(" %2d",get_len(b2[i]));
      /* printf("\n#"); */ testsym(a);
      get_nb(a,MaxDist2,nb);		/* SLOW!!! */
      printf(" Ig %d",geom_imperfection(MaxDist2,nb,1,((d==3)?BCC:SC),v));
      printf(" idx %d",u2i(c)); /* v1.5.4 */
    }
    if(vvv&4){ printf("# square lattice\n"); latticesquare(); }
    if(vvv&8){ printf("# BCC lattice\n");    latticebcc(); }
    printf("\n");
    return 0;
}
